"""
Inherit Description classes and provide a `run` method
that eventually returns a list of errors.
"""
from . import error, description, Executable


def filter_value(value, options: dict = {}):
    if 'uppercase' in options:
        value = value.upper()
    if 'lowercase' in options:
        value = value.lower()
    if 'trim' in options:
        value = value.strip()
    return value


def match(options, value, where=None):
    issues = []
    for case in options:
        value = filter_value(value, case)

        if 'regex' in case:
            if not value.grep(case['regex']):
                issues += [error.InvalidRegex(value, case['regex'], on=where)]

        if 'contains' in case:
            if case['contains'] not in value:
                issues += [
                    error.InvalidContains(
                        value, case['contains'], on=where)]
        if 'equals' in case:
            if case['equals'] != value:
                issues += [
                    error.InvalidEquals(value, case['equals'], on=where)]

    return issues


class TestCase(description.Test):
    def run(self):
        if not isinstance(self.executable, Executable):
            raise ValueError('Not a valid executable')

        output = self.executable.run(*self.args, stdin=self.stdin)

        return [
            *self._check_exit_status(output),
            *self._check_stdout(output),
            *self._check_stderr(output)
        ]

    def _check_exit_status(self, output):
        if 'exit' not in self:
            return []
        if (output.exit_status != self.exit):
            return [
                error.InvalidExitStatus(
                    output.exit_status, self.exit
                )
            ]
        return []

    def _check_stdout(self, output):
        if self.stdout is None:
            return []
        return match(self.stdout, output.stdout, where='stdout')

    def _check_stderr(self, output):
        if self.stderr is None:
            return []
        return match(self.stderr, output.stderr, where='stderr')


class TestGroup(description.Group):
    def run(self):
        return [u.run() for u in self]


class TestSuite(description.Tests):
    _unit_class = TestCase
    _group_class = TestGroup

    def run(self):
        return [u.run() for u in self]



from unittest import TestCase
from voluptuous.error import MultipleInvalid

import baygon.schema as schema

class TestSchema(TestCase):
    def test_minimal(self):
        schema.schema({
            'version': 1,
            'tests': [
                {'exit': 0}
            ]
        })

    def test_wrong_version(self):
        self.assertRaises(MultipleInvalid, schema.schema, {
            'version': 2,
            'tests': [
                {'exit': 0}
            ]
        })

    def test_filters(self):
        schema.schema({
            'version': 1,
            'filters': {
                'uppercase': True
            },
            'tests': [
            ]
        })

    def test_test_contains(self):
        schema.schema({'tests': [{'args': ['--version'], 'stderr': [{'contains': 'Version'}]}]})

    def test_empty_filters(self):
        s = schema.schema({
            'version': 1,
            'tests': []
        })
        self.assertIn('filters', s)

    # Find a way to make it exclusive...
    # def test_filters_exclusive(self):
    #     self.assertRaises(MultipleInvalid, schema.schema, {
    #         'version': 1,
    #         'filters': {
    #             'uppercase': True,
    #             'lowercase': True
    #         },
    #         'tests': [
    #         ]
    #     })



import os
from unittest import TestCase

from baygon import Executable
from baygon.executable import GreppableString

dir_path = os.path.dirname(os.path.realpath(__file__))

class TestExecutable(TestCase):
    def test_not_a_file(self):
        self.assertRaises(ValueError, Executable, 'not-a-file')

    def test_stdout(self):
        e = Executable(dir_path + '/dummy.py')
        output = e.run()
        print(output)
        self.assertEquals(output.stdout, "an apple\n")

    def test_stderr(self):
        e = Executable(dir_path + '/dummy.py')
        output = e.run()
        print(output)
        self.assertEquals(output.stderr, "an orange\n")

    def test_stdin(self):
        e = Executable('/usr/bin/cat')
        test_string = 'Live as if you were to die tomorrow'
        output = e.run(stdin=test_string)
        print(output)
        self.assertEquals(output.stdout, test_string)

    def test_exit_status(self):
        e = Executable(dir_path + '/dummy.py')
        output = e.run()
        print(output)
        self.assertEquals(output.exit_status, 42)

    def test_args(self):
        e = Executable(dir_path + '/args.py')
        test_string = 'foobar'
        output = e.run(2, test_string)
        print(output)
        self.assertEquals(output.stdout, test_string + '\n')

    def test_grep(self):
        s = GreppableString('Live as if you were to die tomorrow')
        u = s.grep(r'\b[aeiouy]\w{2}\b')
        print(u)
        self.assertEquals(u, ['you'])



import click
import time
import logging
import os

from . import TestSuite, TestGroup, TestCase, Executable


def display_pad(pad=0):
    if pad == 0:
        click.secho('  ', nl=False)
    else:
        click.secho('.' * pad, nl=False, dim=True)


def test_name_length(test):
    pad = '  ' * (len(test._id) - 1)
    return len(f'{pad}Test {test.id}: {test.name}')


def display_test_name(test):
    pad = '  ' * (len(test._id) - 1)
    click.secho(f'{pad}Test {test.id}: ', nl=False, bold=True)
    click.echo(f'{test.name}', nl=False)


class OneLineExceptionFormatter(logging.Formatter):
    def formatException(self, exc_info):
        result = super().formatException(exc_info)
        return repr(result)

    def format(self, record):
        result = super().format(record)
        if record.exc_text:
            result = result.replace("\n", "")
        return result


class Runner:
    def __init__(self, verbose, executable=None, **kwargs):
        if verbose > 3:
            self._init_logger("DEBUG")

        self.verbose = verbose
        self.executable = executable
        self.limit = -1 if 'limit' not in kwargs else kwargs['limit']

        self.test_suite = TestSuite(executable=Executable(self.executable))

    def _init_logger(self, loglevel):
        handler = logging.StreamHandler()
        formatter = OneLineExceptionFormatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        root = logging.getLogger()
        root.setLevel(os.environ.get("LOGLEVEL", loglevel))
        root.addHandler(handler)

    def run(self):
        start_time = time.time()
        self.align_column = self._max_length(self.test_suite) + 10

        self.failures = 0
        self.successes = 0

        self._traverse_group(self.test_suite)

        click.secho('\nRan %d tests in %ss.' % (
            self.successes + self.failures,
            round(time.time() - start_time, 2)
        ), bold=True)

        if self.failures > 0:
            click.secho('%d failed, %d passed (%d%% ok).' % (
                self.failures, self.successes,
                100 - round(self.failures / (self.failures + self.successes) * 100, 2)
            ), fg='yellow', bold=True)
            click.secho('\nfail.', fg='red', bold=True)
        else:
            click.secho('\nok.', fg='green')

        return self.failures

    def _max_length(self, tests):
        length = 0
        for test in tests:
            if isinstance(test, TestSuite):
                length = max(test_name_length(test), length)
                length = max(self._max_length(test), length)
            else:
                length = max(test_name_length(test), length)
        return length

    def _traverse_group(self, tests):
        for test in tests:
            if self.limit > 0 and self.failures > self.limit:
                break

            if isinstance(test, TestGroup):
                display_test_name(test)
                click.echo('')
                self._traverse_group(test)
            elif isinstance(test, TestCase):
                display_test_name(test)
                issues = test.run()
                display_pad(self.align_column - test_name_length(test))
                if not len(issues):
                    self.successes += 1
                    click.secho(' PASSED', fg='green')
                else:
                    self.failures += 1
                    click.secho(' FAILED', fg='red', bold=True)
                    for issue in issues:
                        click.secho('  ' * len(test._id) + '- ' + str(issue),
                                    fg='magenta', bold=True)
        return self.failures


@click.command()
@click.argument('executable', required=False, type=click.Path(exists=True))
@click.option('-v', '--verbose', count=True)
@click.option('-l', '--limit', type=int, default=-1)
def cli(verbose=0, executable=None, **kwargs):
    runner = Runner(verbose, executable, **kwargs)
    failures = runner.run()
    click.echo('')
    return failures


if __name__ == '__main__':
    cli()



__version__ = '0.1.5'

__author__ = 'Yves Chevallier'
__description__ = 'Functional tests for teaching activities'
__email__ = 'yves.chevallier@heig-vd.ch'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 HEIG-VD'

from .executable import Executable
from .test import TestCase, TestSuite, TestGroup

__all__ = ['TestCase', 'TestGroup', 'TestSuite', 'Executable']



"""
Read and validate test file.
"""
import yaml
import json

import os
from collections.abc import Sequence

from . import schema, Executable


def check_executable(executable: Executable, filters=None):
    if executable and not isinstance(executable, Executable):
        raise AttributeError('Not an instance of Executable')

    if executable and filters:
        executable.filters = filters

    return executable


def find_testfile(path=None):
    """Recursively find the tests description file."""
    if not path:
        path = os.path.dirname(os.path.realpath('.'))

    if not os.path.isdir(path):
        raise ValueError(f"Path name '{path}' is not a directory")

    for filename in ['baygon', 't', 'test', 'tests']:
        for ext in ['json', 'yml', 'yaml']:
            f = os.path.join(path, f"{filename}.{ext}")
            if os.path.exists(f):
                return f

    # Recursively search in parent directories
    if os.path.dirname(path) != path:  # Test if root directory
        return find_testfile(os.path.dirname(path))


def load(filename):
    """Load a configuration file (can be YAML or JSON)."""
    def loadYaml(filename):
        with open(filename) as fp:
            return yaml.load(fp, Loader=yaml.FullLoader)

    def loadJson(filename):
        with open(filename) as fp:
            return json.load(fp)

    extension = os.path.splitext(filename)[1]
    if extension in ['.yml', '.yaml']:
        return loadYaml(filename)
    if extension in ['.json']:
        return loadJson(filename)
    raise ValueError(f'Unknown extension: {extension}')


class WithId:
    @property
    def id(self):
        return '.'.join(map(str, self._id))

    def _get_id(self, *k: int):
        return list(self._id) + list(k)


class TestSequence(Sequence):
    def __len__(self):
        return len(self._tests)

    def __repr__(self):
        return self.__class__.__name__ + '(' + repr(self._tests) + ')'

    def __getitem__(self, item):
        return self._tests[item]


class Test(dict, WithId):
    """ Functional test descriptior. """
    def __init__(self, *args, executable: Executable = None, id=[]):
        super(Test, self).__init__(*args)
        self.__dict__ = self
        self._id = id
        self.executable = check_executable(executable)

    def __repr__(self):
        return self.__class__.__name__ + '(' + super().__repr__() + ')'


class Group(TestSequence, WithId):
    """ Group of functional tests optionally identified by a name. """
    def __init__(self, tests, name: str = '', executable: Executable = None, id: list = []):
        self._tests = tests
        self.name = name
        self.executable = check_executable(executable)
        self._id = id


class Tests(TestSequence, WithId):
    _group_class = Group
    _unit_class = Test

    def __init__(self, data=None, path=None, executable: Executable = None, id=[]):
        if not isinstance(data, dict):
            data = self._load(path)

        data = schema.schema(data)  # Validate
        tests = data.pop('tests')
        self.__dict__ = data
        self.filename = path
        self.executable = check_executable(executable, self.filters)
        self._id = id
        self._tests = list(self._build(tests, self._id))

    def _load(self, path=None):
        if not path:
            path = os.path.realpath('.')

        if path and not os.path.isfile(path):
            old_path = path
            path = find_testfile(path)
            if not path:
                raise(ValueError(
                    f"Couldn't find and configuration file in '{old_path}'"))

        return load(path)

    def _build(self, tests, id=[]):
        for index, test in enumerate(tests, start=1):
            new_id = id + [index]
            kwargs = {'id': new_id, 'executable': self.executable}
            if 'tests' in test:
                yield self._group_class(
                    list(self._build(test['tests'], id=new_id)),
                    name=test['name'], **kwargs)
            else:
                yield self._unit_class(test, **kwargs)
